/* 
 * Copyright (c) 2000-2003, Andy Gilligan
 * All rights reserved.
 * 
 * This software is not subject to any export provision of the United
 * States Department of Commerce, and may be exported to any country
 * or planet.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice immediately at the beginning of the file, this list of
 *    conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 * 
 * 3. The name of the author may not be used to endorse or promote
 *    products derived from this software without specific prior written
 *    permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 */ 

/* $Id: buildcmd.c,v 1.5.2.1 2003/06/11 03:07:17 ajg Exp $ */

#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <syslog.h>
#include <fcntl.h>
#include "defs.h"
#include "mempool.h"

extern int verbose;
extern int insert;
extern int ipv6;
extern int test_only;

extern void *palloc(struct list *pool, int sz);
extern void pfree(struct list *pool);

extern struct list mpool;

struct keytable {
	char *format;
	char *keyword[2];
	int need_arg;
};

struct keytable kt[] = {
	{ "%s -m state --state ESTABLISHED,RELATED", 	{"keep",	""},		1 },
	{ "%s -m state --state ESTABLISHED,RELATED", 	{"keep-state",	"keepstate"},	0 },
	{ "%s -m state --state INVALID",		{"invalid",	""},		0 },
	{ "%s -f",					{"fragment",	"frag"},	0 },
	{ "%s -s %s",					{"from",	"src"},		1 },
	{ "%s -d %s",					{"to",		"dst"},		1 },
	{ "%s --reject-with %s",			{"with",	""},		1 },
	{ "%s -m quota --quota %s", 			{"quota",	""},		1 },
	{ "%s -m random --average %s",			{"random",	""},		1 },
	{ "%s --tcp-flags ALL %s",			{"flags",	"tcpflags"},	1 },
	{ "%s -m ttl --ttl-eq %s",			{"ttl",		"ttl-eq"},	1 },
	{ "%s -m ttl --ttl-gt %s",			{"ttl-gt",	""},		1 },
	{ "%s -m ttl --ttl-lt %s",			{"ttl-lt",	""},		1 },
	{ "%s -m mark --mark %s",			{"mark",	""},		1 },
	{ "%s -m mac --mac-source %s",			{"mac",		""},		1 },
	{ "%s -p %s",					{"proto",	"protocol"},	1 },
	{ "%s -m iplimit --iplimit-above %s",		{"limit",	"limit-above"},	1 },
	{ "%s -m iplimit --iplimit-mask %s",		{"limit-mask",	""}, 		1 },
	{ "%s -m string --string %s",			{"string",	""}, 		1 },
	{ "%s -m pkttype --pkt-type %s",		{"pkttype",	"pkt-type"},	1 },
	{ "%s -m mport --dports %s",			{"ports",	"dports"},	1 },
	{ "%s -m mport --sports %s",			{"sports",	""},		1 },
	{ "%s --dport %s",				{"port",	"dport"},	1 },
	{ "%s --sport %s",				{"sport",	""},		1 },
	{ "%s -m icmp --icmp-type %s",			{"icmptype",	"icmp-type"},	1 },
	{ "%s -m owner --uid-owner %s",			{"uid",		"user"},	1 },
	{ "%s -m owner --gid-owner %s",			{"gid",		"group"},	1 },
	{ "%s -m owner --pid-owner %s",			{"pid",		""},		1 },
	{ "%s -m owner --sid-owner %s",			{"sid",		""},		1 },
	{ "%s -m owner --cmd-owner %s",			{"cmd",		""},		1 },
	{ "%s -m length --length %s",			{"length",	""},		1 },
	{ "%s -m time --timestart %s",			{"timestart",	"time-start"},	1 },
	{ "%s -m time --timestop %s",			{"timestop",	"time-stop"},	1 },
	{ "%s -m time --days %s",			{"days",	"time-days"},	1 },
	{ "%s -m unclean",				{"unclean",	""},		0 },
	{ 0, {0,0}, 0 }
};

int buildcmd(char **opt)
{
	int c, rc = 0;
	char dir = 'i';
	char *cmd = 0, *tmp = 0, *iptables = 0;
	char *iptcmd = ipv6 ? IP6TABLES : IPTABLES;

	// open up syslog, so we can use it later.
	openlog("agt", LOG_ODELAY, LOG_USER);

	// allocate some buffers... ;)
	if ((iptables = palloc(&mpool, BUF_MAX)) == 0) {
		perror("palloc()");
		goto cleanup;
	}

	if ((cmd = palloc(&mpool, BUF_MAX)) == 0) {
		perror("palloc()");
		goto cleanup;
	}

	if ((tmp = palloc(&mpool, BUF_MAX)) == 0) {
		perror("palloc()");
		goto cleanup;
	}

	// just make sure the person running us has ACTUALLY READ the config!
	if (strcmp(opt[0], "i-have-not-edited-the-configuration") == 0) {
		rc = 200;
		goto cleanup;
	}

	// find out if we should run either 'iptables' or 'iptables -v'
	// or even 'echo iptables'
	snprintf(iptables, BUF_MAX, "%s%s%s", test_only ? "echo ":"", iptcmd,
	                   verbose ? " -v":"");

	if (strcmp(opt[0], "new") == 0) {
		snprintf(cmd, BUF_MAX, "%s -N %s", iptables, opt[1]);
		add2log(cmd);
		rc = system(cmd);
		goto cleanup;
	}

// The Joy of NAT ;)
// FIXME: support for NETMAP targets would be nice
	if (strcmp(opt[0], "nat") == 0) {
		
		{
			// we need to make sure forwarding is enabled, or there
			// is little point in NAT.
			int f;
			if ((f = open("/proc/sys/net/ipv4/ip_forward", O_WRONLY)) != -1) {
				(void) write(f, "1", 1);
				(void) close(f);
			}
		}
		
		// any case of nat/masquerading we support, will always have
		// a POSTROUTING entry.
		snprintf(cmd, BUF_MAX, "%s -t nat -%s POSTROUTING", iptables, insert ? "I":"A");

		if (strcmp(opt[1], "from") == 0) {
			// 'iptables -t nat -A POSTROUTING -s 10.0.0.1'
			snprintf(tmp, BUF_MAX, "%s -s %s", cmd, opt[2]);

			if (opt[3] && opt[4] && (strcmp(opt[3], "to") == 0)) {
				// 'iptables -t nat -A POSTROUTING -s 10.0.0.1 -d ! 10.0.0.1 -j SNAT --to-source 1.2.3.4'
				snprintf(cmd, BUF_MAX, "%s -d ! %s -j SNAT --to-source %s", tmp, opt[2], opt[4]);
				
				add2log(cmd);
				// cleanup if system() returns an error
				if ((rc = system(cmd)))
					goto cleanup;
				
				// if we want NAT (not MASQ), add a return rule to map incoming connections.
				// 'iptables -t nat -A PREROUTING -d 1.2.3.4 -j DNAT --to-destination 10.0.0.1'
				snprintf(cmd, BUF_MAX, "%s -t nat -%s PREROUTING "
				                       "-d %s -j DNAT --to-destination %s", iptables,
				                       insert ? "I":"A", opt[4], opt[2]);
				add2log(cmd);
				// cleanup if system() returns an error
				if ((rc = system(cmd)))
					goto cleanup;
			} else {
				// just simple masquerading then :)
				// 'iptables -t nat -A POSTROUTING -s 10.0.0.1 -j MASQUERADE'
				snprintf(cmd, BUF_MAX, "%s -j MASQUERADE", tmp);
				add2log(cmd);
				// cleanup if system() returns an error
				if ((rc = system(cmd)))
					goto cleanup;
			}
		}
		// none of the tests below are relevant if we are doing NAT.
		goto cleanup;
	}

// FIXME: this needs to be integrated with the nat tests above

	// begin our portforwarding tests
	if (strcmp(opt[0], "portfw") == 0) {
		int i;
		char *proto = "tcp";
		char *to = 0, *port = 0;

		for (i = 0; opt[i]; i++) {
			if (strcmp(opt[i], "proto") == 0) {
				proto = opt[++i];
				continue;
			}
			
			if (strcmp(opt[i], "to") == 0) {
				to = opt[++i];
				continue;
			}
			
			if (strcmp(opt[i], "port") == 0) {
				port = opt[++i];
				continue;
			}
		}

		// we can continue if we have a port and a destination.
		// we default to TCP if no protocol is specified.
		if (port && to) {
			snprintf(cmd, BUF_MAX, "%s -t nat -%s PREROUTING -p %s --dport %s -j DNAT --to %s",
			                       iptables, insert ? "I":"A", proto, port, to);
			add2log(cmd);
			rc = system(cmd);
		}
		goto cleanup;
	}

	{
		int x = 0;

		// cmd = 'iptables'
		snprintf(cmd, BUF_MAX, "%s", iptables);

		if (strcmp(opt[x], "default") == 0) {
			// ok, we have to set the default policy
			strncat(cmd, " -P ", BUF_MAX);
			x++;
		}
		else
			// we aren't setting a default policy, so we need
			// to say we're adding a rule. (or inserting)
			strncat(cmd, insert ? " -I ":" -A ", BUF_MAX);

		if (strcmp(opt[x], "input")   == 0) strncat(cmd, "INPUT",   BUF_MAX); else
		if (strcmp(opt[x], "output")  == 0) { strncat(cmd, "OUTPUT",  BUF_MAX); dir = 'o'; } else
		if (strcmp(opt[x], "forward") == 0) strncat(cmd, "FORWARD", BUF_MAX); else
			strncat(cmd, opt[x], BUF_MAX);

		x++;

		// we should have more args, bail if we don't
		if (opt[x] == 0) {
			rc = 1;
			goto cleanup;
		}

		// when x == 1, it means that we haven't matched the
		// 'default' test above -- we need to add '-j'
		if (x == 1)
			strncat(cmd, " -j", BUF_MAX);

		// common targets are specified as lower case in the config
		if (strcmp(opt[x], "accept") == 0) strncat(cmd, " ACCEPT", BUF_MAX); else
		if (strcmp(opt[x], "mirror") == 0) strncat(cmd, " MIRROR", BUF_MAX); else
		if (strcmp(opt[x], "reject") == 0) strncat(cmd, " REJECT", BUF_MAX); else
		if (strcmp(opt[x], "log")    == 0) strncat(cmd, " LOG",    BUF_MAX); else
		if (strcmp(opt[x], "drop")   == 0) strncat(cmd, " DROP",   BUF_MAX); else {
			// didn't match our standard targets, must be a custom target
			strncat(cmd, " ", BUF_MAX);
			strncat(cmd, opt[x], BUF_MAX);
		}
	}

	// start loop skipping first two keywords
	// eg. 'input' + 'accept'
	for (c = 2; opt[c]; c++) {
		int i;

// just to make our code a bit cleaner ;)
#define match(x) (strcmp(opt[c], (x)) == 0)
#define NEED_ARG (kt[i].need_arg)
		
		for (i = 0; kt[i].format; i++) {
			if (match(kt[i].keyword[0]) || match(kt[i].keyword[1])) {
				
				if (NEED_ARG && opt[c + 1] == 0) {
					printf("agt: keyword '%s' requires an argument.\n", opt[c]);
					goto cleanup;
				}

				if (NEED_ARG)
					snprintf(tmp, BUF_MAX, kt[i].format, cmd, opt[++c]);
				else
					snprintf(tmp, BUF_MAX, kt[i].format, cmd);

				strncpy(cmd, tmp, BUF_MAX);
				
				// break out of THIS loop (we're nested)
				break;
			}
		}

		// did we find a keyword?
		// if we didn't, kt[i].format will be null.
		// (we couldn't add 'continue' to the above test, as we were
		// inside a nested loop)
		if (kt[i].format)
			continue;

		// this could be put into the main tests array, but we need
		// to specify either -i or -o for the interface direction.
		if ((strcmp(opt[c], "interface") == 0) || (strcmp(opt[c], "on") == 0)) {
			if (opt[c+1]) {
				snprintf(tmp, BUF_MAX, "%s -%c %s", cmd, dir, opt[++c]);
				strncpy(cmd, tmp, BUF_MAX);
			} else {
				printf("agt: keyword '%s' requires an argument.\n", opt[c]);
				goto cleanup;
			}
			continue;
		}
	}
	
	// right, send our command to the shell, and log it
	add2log(cmd);
	rc = system(cmd);

cleanup:
	// close syslog
	closelog();
	
	// if rc != 0, then we have an error
	return rc;
}

