/* 
 * Copyright (c) 2000-2003, Andy Gilligan
 * All rights reserved.
 * 
 * This software is not subject to any export provision of the United
 * States Department of Commerce, and may be exported to any country
 * or planet.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice immediately at the beginning of the file, this list of
 *    conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 * 
 * 3. The name of the author may not be used to endorse or promote
 *    products derived from this software without specific prior written
 *    permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 */

/* $Id: main.c,v 1.6.2.4 2003/06/11 03:06:38 ajg Exp $ */

#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/utsname.h>
#include <getopt.h>
#include <time.h>
#include "defs.h"
#include "mempool.h"

extern char *sws(char *str);
extern char *getline(char *line, int max, FILE *fp);
extern void parse(char **buf, char *line, int maxent);
extern int buildcmd(char **opt);

extern void *palloc(struct list *pool, int sz);
extern void pfree(struct list *pool);

int verbose = 0;
int ipv6 = 0;
int test_only = 0;
int append = 0;
int insert = 0;
int quiet = 0;

struct list mpool = { 0, 0 };

char *config = AGT_RULES;

char *short_opt = "6f:hqtvV";
struct option long_opt[] =
{
	{ "ipv6",	0, 0, '6' },
	{ "file",	1, 0, 'f' },
	{ "help",	0, 0, 'h' },
	// FIXME: rules are inserted in reverse order
	//        until this is fixed, inserting is disabled.
	//{ "insert",	0, 0, 'i' },
	// appending rules is the default, doesn't make sense to
	// have the option if inserting is disabled
	//{ "append",	0, 0, 'a' },
	{ "quiet",	0, 0, 'q' },
	{ "test",	0, 0, 't' },
	{ "verbose",	0, 0, 'v' },
	{ "version",	0, 0, 'V' },
	{ 0, 0, 0, 0 }
};	

void clearfw(void)
{
	int c;
	char *list[] = { "filter", "nat", "mangle", 0 };
	char *cmd = 0;
	char *ipt = "iptables &>/dev/null -t";

	if ((cmd = calloc(1, BUF_MAX))) {
		for (c = 0; list[c]; c++) {
			snprintf(cmd, BUF_MAX, "%s %s -F", ipt, list[c]); (void) system(cmd);
			snprintf(cmd, BUF_MAX, "%s %s -X", ipt, list[c]); (void) system(cmd);
		}
		free(cmd);
	}
}

int main(int argc, char **argv)
{
	FILE *fp = 0;
	char *line = 0, **opt = 0;
	int c, rc = EXIT_SUCCESS;
	struct utsname u;
	
	if (uname(&u)) {
		perror("uname()");
		exit(EXIT_FAILURE);
	}

#ifdef LINUX_ONLY
	// Linux would be nice...
	if (strcmp("Linux", u.sysname)) {
		printf("AGT can only be run on Linux systems.\n");
		exit(EXIT_FAILURE);
	}
#endif
	while ((c = getopt_long(argc, argv, short_opt, long_opt, 0)) != -1) {
		switch(c) {
			case '?':
			case ':':
				exit(EXIT_FAILURE);

			case '6':
				ipv6 = 1;
				break;

			case 'a':
				append = 1;
				break;

			case 'f':
				if (optarg)
					config = optarg;
				break;

			case 'h':
				printf("USAGE: agt [OPTION]...\n\n"
				       //"  -a, --append       append rules to bottom of firewall table\n"
				       "  -f, --file=CONF    use alternate configuration file\n"
				       "  -h, --help         display this help\n"
				       //"  -i, --insert       insert rules at top of firewall table\n"
				       "  -q, --quiet        suppress all messages, except for errors\n"
				       "  -t, --test         do not apply rules, just test configuration\n"
				       "  -v, --verbose      explain what is being done\n"
				       "  -V, --version      output version information and exit\n\n");
				exit(EXIT_SUCCESS);
				break;
			
			case 'i':
				insert = 1;
				break;
				
			case 'q':
				quiet = 1;
				break;

			case 't':
				test_only = 1;
				break;

			case 'v':
				verbose = 1;
				break;

			case 'V':
			{
				char year[10];
				time_t t;
				time(&t);
				strftime(year, 9, "%Y", localtime(&t));

				printf("AGT " VERSION "\n"
				       "Written by Andy Gilligan.\n\n"
				       "Copyright (c) 2000-%s, Andy Gilligan\n"
				       "This is free software; see the source for copying conditions.  There is NO\n"
				       "warranty; not even for MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.\n",
				       year);
			}
				exit(EXIT_SUCCESS);
				break;

			default:
				break;
		}
	}

	if (verbose && quiet) {
		// MAKE UP YOUR MIND!
		printf("Options 'quiet' and 'verbose' cannot be specified together!\n");
		exit(EXIT_FAILURE);
	}

#ifdef ROOT_ONLY
	// we'll just bail now if we're not root
	// (saves on lots of 'Permission denied' responses from iptables!)
	if (getuid()) {
		printf("You must have root privileges to run AGT.\n");
		exit(EXIT_FAILURE);
	}
#endif
	if ((fp = fopen(config, "r"))) {
		if ((line = palloc(&mpool, BUF_MAX * sizeof(line)))) {
			if ((opt = palloc(&mpool, OPT_MAX * sizeof(opt)))) {
				int n = 0;

				// don't clear or flush anything if we're just testing!
				// (or if we are appending or inserting rules)
				if (test_only == 0 && append == 0 && insert == 0)
					clearfw();

				// if we're not being quiet, verbose, or just testing
				// output our status.  this will be finished up later.
				if (quiet == 0 && verbose == 0 && test_only == 0) {
					printf("AGT running... ");
					(void) fflush(stdout);
				}

				while(getline(line, BUF_MAX, fp)) {
					parse(opt, line, OPT_MAX);
					
					if ((rc = buildcmd(opt))) {

						if (rc == 200) {
							// hmm, didn't read the config, huh?
							printf("PLEASE read and edit the configuration before running AGT!\n");
							rc = 1;
							goto cleanup;
						}

						// we have a non-zero return from buildcmd(), so something
						// went wrong, tell the user, and say which line it was.
						printf("\nProblem with configuration, last line was:\n");
						
						for (c = 0; opt[c]; c++)
							printf("%s ", opt[c]);
						
						printf("\n\nMaybe you lack the proper kernel module?\n");
						printf("Have you specified the correct protocol?\n");

						// clear the firewall only if we aren't testing, inserting,
						// or appending rules, and only if we have modified it
						// (this prevents some lockouts when working remotely)
						if (test_only == 0 && append == 0 && insert == 0 && n)
							clearfw();
						
						goto cleanup;
					} else
						// increment our 'rules added' counter.
						n++;
				}

				// if we're not being quiet, verbose, or just testing
				// output the number of rules we've added.
				if (quiet == 0 && verbose == 0 && test_only == 0)
					printf(" %d rule%s added\n", n, n == 1 ? "":"s");

				// hmm, not a single rule found... oh well...
				if (opt[0] == 0)
					printf("No usable rules found! - Check %s!\n", config);
			} else {
				perror("palloc()");
				rc = EXIT_FAILURE;
			}
		} else {
			perror("palloc()");
			rc = EXIT_FAILURE;
		}
	} else {
		perror(config);
		rc = EXIT_FAILURE;
	}

cleanup:
	if (fp) (void) fclose(fp);
	pfree(&mpool);
	exit(rc);
}

